#ifndef __PARSLIB_H
#define __PARSLIB_H 

#include <string>
#include <sstream>
#include <iostream>
#include <vector>
#include "../util/Show.h"
#include "../util/Tuple.h"
#include "../util/Either.h"

using namespace std;
using namespace name::panitz::util;

namespace name {namespace panitz{ namespace parser{

template <class resultType>
class CAF {

  private:
    resultType(*f)();

  public:
    resultType result;

    resultType operator()(){
      if (result==NULL) result=f();
      return result;
    }

    CAF(){}
    CAF(resultType x):result(x){}
    CAF(resultType(*f)()){this->f=f;result=NULL;}
 
    virtual string toString(){return "CAF(..)";}
};

template <class resultType>
CAF<resultType> caf(resultType(*f)()){return CAF<resultType>(f);}

template <class resultType>
CAF<resultType> caf(resultType x){return CAF<resultType>(x);}

template <typename resultType,typename tokenType>
class ParsResult{

  private: 
    resultType result;
    vector<tokenType*> furtherToken;

  public: 
    ParsResult(resultType result,vector<tokenType*>& furtherToken) {
      this->result=result;
      this->furtherToken=furtherToken;
      ParsResultCount=ParsResultCount+1;
    }
    ParsResult(vector<tokenType*>& furtherToken) {
      this->furtherToken=furtherToken;
      ParsResultCount=ParsResultCount+1;
    }
    ~ParsResult(){ParsResultCount=ParsResultCount-1;}

    resultType getResult(){return result;}
    vector<tokenType*>& getFurtherToken(){return furtherToken;}

    virtual bool failed(){return false;}

    virtual string toString(){
      return "ParsResult("+show(result)
                      +","+show(getFurtherToken())+")";}
};

static int ParsResultCount=0;

template <typename resultType,typename tokenType>
class Fail:public ParsResult<resultType,tokenType>{
  public:
    Fail(vector<tokenType*>& furtherToken)
        :ParsResult<resultType,tokenType>(furtherToken){
    }
    virtual bool failed(){return true;}
    virtual string toString(){
     return "Fail("+show(getFurtherToken())+")";}
};

template <typename resultType,typename tokenType>
class Parser{
  public:
    virtual ParsResult<resultType,tokenType>* 
                  parse(vector<tokenType*>& xs)=0;

    virtual ~Parser(){}; 
    virtual std::string toString()=0; 
};

template <typename tokenType>
class GetToken: public Parser<tokenType*,tokenType>{

  private:
     tokenType* token;
     bool(*eq)(tokenType*,tokenType*);

  public:
    GetToken(tokenType* token,bool(*eq)(tokenType*,tokenType*))
     {this->token=token;this->eq=eq;}

    ParsResult<tokenType*,tokenType>* parse(vector<tokenType*>& xs){
      if (!xs.empty() && eq(token,xs[0])){
        tokenType* tok = xs[0];
        vector<tokenType*> restToken 
         = vector<tokenType*>(xs.begin()+1,xs.end());
        ParsResult<tokenType*,tokenType>*result
          =new ParsResult<tokenType*,tokenType>(tok,restToken) ;
        return result;
      }
      return  new Fail<tokenType*,tokenType>(xs);
    }

    virtual std::string toString(){
      return "GetToken("+show(token)+")";
    }
};

template <typename tokenType>
Parser<tokenType*,tokenType>* 
       getToken(tokenType* token,bool(*eq)(tokenType*,tokenType*)){
  return new GetToken<tokenType>(token,eq);}

template 
  <typename resultType1,typename resultType2,typename tokenType>

class Seq:public Parser<Pair<resultType1,resultType2>*,tokenType >{

  private:
     CAF<Parser<resultType1,tokenType>*> p1;
     CAF<Parser<resultType2,tokenType>*> p2;

  public:
    Seq(CAF<Parser<resultType1,tokenType>*> p1
       ,CAF<Parser<resultType2,tokenType>*> p2){
      this->p1=p1;
      this->p2=p2;
    }

    virtual ParsResult<Pair<resultType1,resultType2>*,tokenType>*
                                     parse(vector<tokenType*>& xs){

      ParsResult<resultType1,tokenType>* res1 = p1()->parse(xs);

     if (!res1->failed()) {
        vector<tokenType*> further = res1->getFurtherToken();
        resultType1 r1 = res1->getResult();         

        ParsResult<resultType2,tokenType>* res2 
                                  = p2()->parse(further);

        if (!res2->failed()){
          resultType2 r2 = res2->getResult();  
          vector<tokenType*> further = res2->getFurtherToken();

	  delete res1;
	  delete res2;
          return 
          new ParsResult<Pair<resultType1,resultType2>*,tokenType>
             (new Pair<resultType1,resultType2>(r1,r2),further);
        }
        delete res2;
      }

      delete res1;
      return new Fail<Pair<resultType1,resultType2>*,tokenType>(xs);
    }

    virtual std::string toString(){
	return "Seq(..)";
    }
};

template <typename a,typename b,typename c>
Parser<Pair<a,b>*,c>* seq(CAF<Parser<a,c>*>p1,CAF<Parser<b,c>*> p2){
  return ((Parser<Pair<a,b>*,c>*)new Seq<a,b,c>(p1,p2));
}

template <typename a,typename b,typename c>
Parser<Pair<a,b>*,c>* 
    operator-(CAF<Parser<a,c>*> p1,CAF<Parser<b,c>*> p2){
  return seq(p1,p2);
}

template <typename a,typename b,typename c>
CAF<Parser<Pair<a,b>*,c>*> 
  operator,(CAF<Parser<a,c>*> p1,CAF<Parser<b,c>*> p2){
  return caf(seq(p1,p2));
}

template 
 <typename resultType1,typename resultType2,typename tokenType>
class Alt:public Parser<Either<resultType1,resultType2>*,tokenType>{

  private:
     CAF<Parser<resultType1,tokenType>*> p1;
     CAF<Parser<resultType2,tokenType>*> p2;

  public:
    Alt( CAF<Parser<resultType1,tokenType>*> p1
       , CAF<Parser<resultType2,tokenType>*> p2){
      this->p1=p1;
      this->p2=p2;
    }

    virtual ParsResult<Either<resultType1,resultType2>*,tokenType>*
                                     parse(vector<tokenType*>& xs){

      ParsResult<resultType1,tokenType>* res1 = p1()->parse(xs);

      if (!res1->failed()) {
        vector<tokenType*> further = res1->getFurtherToken();
        resultType1 r1 = res1->getResult();
        delete res1;
        return
        new ParsResult<Either<resultType1,resultType2>*,tokenType> 
           (new Left<resultType1,resultType2>(r1),further);
      }

      delete res1;
      ParsResult<resultType2,tokenType>* res2 = p2()->parse(xs);
      if (!res2->failed()) {
        resultType2 r2 = res2->getResult();
        vector<tokenType*> further = res2->getFurtherToken();
        delete res2;
        return 
        new ParsResult<Either<resultType1,resultType2>*,tokenType> 
           (new Right<resultType1,resultType2>(r2),further);
      }

      delete res2;
      return
       new Fail<Either<resultType1,resultType2>*,tokenType>(xs);
    }

    virtual std::string toString(){
	return "Alt(..)";
    }
};

template <typename a,typename b,typename c>
Parser<Either<a,b>*,c>* 
  alt(CAF<Parser<a,c>*>p1,CAF<Parser<b,c>*> p2){
  return ((Parser<Either<a,b>*,c>*)new Alt<a,b,c>(p1,p2));
}

template <typename a,typename b,typename c>
Parser<Either<a,b>*,c>* 
    operator|(CAF<Parser<a,c>*> p1,CAF<Parser<b,c>*> p2){
  return alt(p1,p2);
}

template <typename X,typename Y,typename tokenType>
class Map:public Parser<Y,tokenType>{

  private:
    Parser<X,tokenType>* p;
    Y(*f)(X);

  public:
    Map(Y(*f)(X),Parser<X,tokenType>* p):f(f),p(p){}

    virtual ParsResult<Y,tokenType>* parse(vector<tokenType*>& xs){

     ParsResult<X,tokenType>* res1 = p->parse(xs);

      if (!res1->failed()){
        X r1 = res1->getResult();
        vector<tokenType*> further = res1->getFurtherToken();
        delete res1;

        return new ParsResult<Y,tokenType>((*f)(r1),further);
      }

      delete res1;
      return new Fail<Y,tokenType>(xs);
    }

    virtual std::string toString(){
	return "Map(..)";
    }
};

template <typename a,typename b,typename c>
Parser<b,c>* map(CAF<Parser<a,c>*> p,  b(*f)(a)){
  return ((Parser<b,c>*)new Map<a,b,c>(f,p()));
}

template <typename a,typename b,typename c>
Parser<b,c>* operator<(CAF<Parser<a,c>*> p,  b(*f)(a)){
  return map(p,f);
}

template <typename a,typename b,typename c>
CAF<Parser<b,c>*> operator<<(CAF<Parser<a,c>*> p,  b(*f)(a)){
  return caf(map(p,f));
}

template <typename a,typename c>
Parser<a,c>* ualt(CAF<Parser<a,c>*> p1,CAF<Parser<a,c>*> p2){
  return map<Either<a,a>*,a,c>(alt(p1,p2),getLeftRight<a>);
}

template <typename a,typename c>
CAF<Parser<a,c>*> 
     operator||(CAF<Parser<a,c>*> p1,CAF<Parser<a,c>*> p2){
  return caf(ualt(p1,p2));
}

template <typename resultType,typename tokenType>
class Repetition0:public Parser<vector<resultType>*,tokenType>{
  private:
    CAF<Parser<resultType,tokenType>*> p;

  public:
    Repetition0(CAF<Parser<resultType,tokenType>*> p):p(p){}

    virtual ParsResult<vector<resultType>*,tokenType>*
                                    parse(vector<tokenType*>& xs){
      vector<tokenType*>ys= xs;
      vector<resultType>* results = new vector<resultType>();
      ParsResult<resultType,tokenType>* res = p()->parse(ys);

      while (!res->failed()){
        results->insert(results->end(),res->getResult());
        ys = res->getFurtherToken();
        delete res;
        res=p()->parse(ys);
      }
      vector<tokenType*> further=res->getFurtherToken();
      delete res;

      return new ParsResult<vector<resultType>*,tokenType>
                                            (results,further);
    }

    CAF<Parser<resultType,tokenType>*> getP(){return p;}

    virtual std::string toString(){
      return "Repetition0("+show(p)+")";
    }
};

template <typename resultType>
vector<resultType>* 
          insertFirst(Pair<resultType,vector<resultType>*>* p){
  vector<resultType>* result = p->snd;
  result->insert(result->begin(),p->fst);
  delete p;
  return result;
}

template <typename resultType,typename tokenType>
class Repetition1:public Repetition0<resultType,tokenType>{
  public:
    Repetition1(CAF<Parser<resultType,tokenType>*> p)
      :Repetition0<resultType,tokenType>(p){}

    virtual ParsResult<vector<resultType>*,tokenType>*
                                    parse(vector<tokenType*>& xs){
      Parser<vector<resultType>*,tokenType>* rep0 
        = new Repetition0<resultType,tokenType>(getP());

      Parser<vector<resultType>*,tokenType>* repP
       =  (getP(),caf(rep0)) < (insertFirst<resultType>);
      return repP->parse(xs);
    }

    virtual std::string toString(){
      return "Repetition1("+show(getP())+")";
    }
};

template <typename a,typename b>
CAF<Parser<vector<a>*,b>*> rep0(CAF<Parser<a,b>*> p){
  return caf((Parser<vector<a>*,b>*)new Repetition0<a,b>(p));
}

template <typename a,typename b>
CAF<Parser<vector<a>*,b>*> rep1(CAF<Parser<a,b>*> p){
  return caf((Parser<vector<a>*,b>*)new Repetition1<a,b>(p));
}

template <typename resultType
         ,typename seperatorType
         ,typename tokenType>
class SeperatedBy:public Parser<vector<resultType>*,tokenType>{
  private:
    CAF<Parser<resultType,tokenType>*> p;
    CAF<Parser<seperatorType,tokenType>*> sep;
  public:
    SeperatedBy(CAF<Parser<resultType,tokenType>*> p
               ,CAF<Parser<seperatorType,tokenType>*> sep
               ):p(p),sep(sep){}
    virtual ParsResult<vector<resultType>*,tokenType>*
                                    parse(vector<tokenType*>& xs){
      vector<tokenType*> ys=xs;
      vector<resultType>* results = new vector<resultType>();
      ParsResult<resultType,tokenType>* res = p()->parse(ys);
      if (res->failed()) {
        delete res;
        return new Fail<vector<resultType>*,tokenType>(ys);
      }

      ys = res->getFurtherToken();
      results->insert(results->end(),res->getResult());

      delete res;
      Parser<Pair<seperatorType,resultType>*,tokenType> * sepPars
       = new Seq<seperatorType,resultType,tokenType>(sep,p);

      ParsResult<Pair<seperatorType,resultType>*,tokenType>* resPair
        = sepPars->parse(ys);

      while (!resPair->failed()){
         results->insert(results->end(),resPair->getResult()->snd);
         ys = resPair->getFurtherToken();
         delete resPair->getResult();
         delete resPair;
         resPair= sepPars->parse(ys);
      }
      vector<tokenType*> further=resPair->getFurtherToken();
      delete resPair;
      delete sepPars;

      return 
      new ParsResult<vector<resultType>*,tokenType>(results,further);
    }

    virtual std::string toString(){
      return "SeperatedBy("+show(p)+","+show(sep)+")";
    }
};

template <typename a,typename b,typename c>
CAF<Parser<vector<a>*,c>*> 
      sepBy(CAF<Parser<a,c>*> p,CAF<Parser<b,c>*> sep){
  return caf((Parser<vector<a>*,c>*) new SeperatedBy<a,b,c>(p,sep));
}

template <typename resultType,typename tokenType>
class Result:public Parser<resultType,tokenType>{
  private:
    resultType result;
  public:
    Result(resultType result):result(result){}

    ParsResult<resultType,tokenType>* parse(vector<tokenType*>& xs){
     return new ParsResult<resultType,tokenType>(result,xs);
    }
    string toString(){return "Result";}
};

template <typename resultType,typename tokenType>
Parser<resultType,tokenType>* result(resultType res){
  return new Result<resultType,tokenType>(res);
}
}}}//namespace end

#endif



